from __future__ import annotations

from mathutils import Vector
from typing import Optional
from math import inf
import bmesh


class Precision:
    DEFAULT_SMALL_NUMBER = 1e-6
    SMALL_NUMBER: float

    @staticmethod
    def set_from_exponent(exponent_positive: int) -> None:
        Precision.SMALL_NUMBER = 10**(-exponent_positive)

    @staticmethod
    def reset() -> None:
        Precision.SMALL_NUMBER = Precision.DEFAULT_SMALL_NUMBER


Precision.reset()


class Plane:
    # index just for the reference
    def __init__(self, center: Vector, normal: Vector, index: int = None) -> None:
        self.center = center
        self.normal = normal
        self.index = index

    # NOTE: for our purpose, just compare the normals of the two planes
    def __eq__(self, other: Plane) -> bool:
        return vectors_equal(self.normal, other.normal, Precision.SMALL_NUMBER)

    # # NOTE: for our purpose, just use the hash of the normal
    # def __hash__(self) -> int:
    #     return id(self.normal)

    def __str__(self) -> str:
        return "Plane: center={} normal={} index={}".format(self.center, self.normal, self.index)

    def __repr__(self) -> str:
        return str(self)

    def distance(self) -> float:
        return point_to_plane_dist_signed(Vector(), self.center, self.normal)


def project_point_on_line(p: Vector, a: Vector, b: Vector,
                          epsilon: float = Precision.SMALL_NUMBER) -> Vector:
    ap = p - a
    ab = b - a
    return a + ap.dot(ab) / max(ab.dot(ab), epsilon) * ab


def is_projected_point_on_segment(p: Vector, a: Vector, b: Vector) -> bool:
    ap = p - a
    ab = b - a
    dot = ab.dot(ap)
    return 0 < dot < ab.length_squared


def project_point_on_plane(p: Vector, plane_origin: Vector, plane_normal: Vector) -> Vector:
    v = p - plane_origin
    dist = v.dot(plane_normal)
    return p - dist * plane_normal


def point_to_plane_dist(p: Vector, plane_origin: Vector, plane_normal: Vector) -> float:
    v = p - plane_origin
    return abs(v.dot(plane_normal))


def point_to_plane_dist_signed(p: Vector, plane_origin: Vector, plane_normal: Vector) -> float:
    v = p - plane_origin
    return -v.dot(plane_normal)


def line_plane_intersection(a: Vector, b: Vector, plane_origin: Vector,
                            plane_normal: Vector, epsilon: float = 1e-6) -> Optional[Vector]:
    u = b - a
    dot = plane_normal.dot(u)

    if abs(dot) > epsilon:
        # The factor of the point between p0 -> p1 (0 - 1)
        # if 'fac' is between (0 - 1) the point intersects with the segment.
        # Otherwise:
        #  < 0.0: behind p0.
        #  > 1.0: infront of p1.
        w = a - plane_origin
        fac = -plane_normal.dot(w) / dot
        u *= fac
        return a + u

    # in case the line is parallel to the plane
    return None


def calculate_median_point(verts: list[Vector]) -> Vector | None:
    if len(verts) == 0:
        return None

    co_sum = Vector([sum([v.x for v in verts]),
                               sum([v.y for v in verts]),
                               sum([v.z for v in verts])])
    return co_sum / len(verts)


class Node:
    @property
    def edges(self):
        return (e for e in self.vert.link_edges if not e.tag)

    def __init__(self, v):
        self.vert = v
        self.length = inf
        self.shortest_path = []


def dijkstra(bm, v_start, v_target=None):
    for e in bm.edges:
        e.tag = False

    d = {v: Node(v) for v in bm.verts}
    node = d[v_start]
    node.length = 0

    visiting = [node]

    while visiting:
        node = visiting.pop(0)

        if node.vert is v_target:
            return d

        for e in node.edges:
            e.tag = True
            length = node.length + e.calc_length()
            v = e.other_vert(node.vert)

            visit = d[v]
            visiting.append(visit)
            if visit.length > length:
                visit.length = length
                visit.shortest_path = node.shortest_path + [e]

        visiting.sort(key=lambda n: n.length)

    return d


def polygon_area(verts: list[Vector]) -> float:
    if len(verts) < 3:
        return 0
    bm = bmesh.new()
    for v in verts:
        bm.verts.new(v)
    result = bm.faces.new(bm.verts).calc_area()
    bm.free()
    return result


def collinear_three_points(a: Vector, b: Vector, c: Vector,
                           threshold: float = Precision.SMALL_NUMBER) -> bool:
    return (b - a).cross(c - a).length < threshold
    # return ((c - a) - (b - a) - (c - b)).length < threshold


clamp = lambda x, l, u: l if x < l else u if x > u else x


def planes_parallel(p1_normal: Vector, p2_normal: Vector, threshold: float = Precision.SMALL_NUMBER) -> bool:
    return abs(p1_normal.normalized().dot(p2_normal.normalized())) > (1.0 - threshold)


def line_plane_parallel(a: Vector, b: Vector, normal: Vector,
                        threshold: float = Precision.SMALL_NUMBER) -> bool:
    return (b - a).normalized().cross(normal.normalized()).length > (1.0 - threshold)


def three_plane_intersection(p1: Plane, p2: Plane, p3: Plane, epsilon: float = Precision.SMALL_NUMBER) -> Vector:
    m1 = Vector((p1.normal.x, p2.normal.x, p3.normal.x))
    m2 = Vector((p1.normal.y, p2.normal.y, p3.normal.y))
    m3 = Vector((p1.normal.z, p2.normal.z, p3.normal.z))
    d = Vector((p1.distance(), p2.distance(), p3.distance()))

    u = m2.cross(m3)
    v = m1.cross(d)
    denom = m1.dot(u)

    if abs(denom) < epsilon:
        raise Exception('Invalid point plane intersection evaluation attempt (perhaps invalid selection?)')

    return Vector((d.dot(u) / denom, m3.dot(v) / denom, -m2.dot(v) / denom))


def vectors_equal(a: Vector, b: Vector, epsilon: float = Precision.SMALL_NUMBER * 100) -> bool:
    return a.dot(b) >= (1.0 - epsilon)


def vectors_equal_abs(a: Vector, b: Vector, epsilon: float = Precision.SMALL_NUMBER * 100) -> bool:
    return abs(a.dot(b)) >= (1.0 - epsilon)
